#define LAYOUT_ANSI 33
#define LAYOUT_ISO  13
#define LAYOUT_JIS  15

static int keyboard_types[] = {
    LAYOUT_ISO,  // Unknown
    LAYOUT_ISO,  // German
    LAYOUT_ISO,  // French
    LAYOUT_JIS,  // Japanese Keyboard
    LAYOUT_ISO,  // US International Keyboard
    LAYOUT_ANSI, // U.S.
    LAYOUT_ISO,  // British
    LAYOUT_ISO,  // Spanish - ISO
    LAYOUT_ISO,  // Swedish - Pro
    LAYOUT_ISO,  // Italian - Pro
    LAYOUT_ISO,  // Canadian - CSA
    LAYOUT_ANSI, // Simplified Chinese Keyboard
    LAYOUT_ISO,  // Danish
    LAYOUT_ISO,  // Belgian
    LAYOUT_ISO,  // Norwegian
    LAYOUT_ANSI, // 2SetHangul
    LAYOUT_ISO,  // Dutch
    LAYOUT_ISO,  // Swiss Multilingual Keyboard
    LAYOUT_ANSI, // Zhuyin Bopomofo
    LAYOUT_ISO,  // Arabic
    LAYOUT_ISO,  // Bulgarian
    LAYOUT_ISO,  // Croatian
    LAYOUT_ISO,  // Czech
    LAYOUT_ISO,  // Greek
    LAYOUT_ISO,  // Hebrew
    LAYOUT_ISO,  // Icelandic
    LAYOUT_ISO,  // Hungarian
    LAYOUT_ISO,  // Polish
    LAYOUT_ISO,  // Portuguese
    LAYOUT_ISO,  // Persian
    LAYOUT_ISO,  // Romanian
    LAYOUT_ISO,  // Russian
    LAYOUT_ISO,  // Slovak
    LAYOUT_ANSI, // Thai
    LAYOUT_ISO,  // Turkish-QWERTY-PC
    LAYOUT_ISO,  // Turkish
    LAYOUT_ISO,  // Ukrainian
    LAYOUT_ISO,  // Turkish-Standard
    LAYOUT_ISO,  // Latin America
};
