/* SPDX-License-Identifier: MIT */

#ifndef LIMITS_H
#define LIMITS_H

#define INT_MAX  (0x7fffffff)
#define UINT_MAX (0xffffffffu)

#define LONG_MAX  (0x7fffffffffffffffl)
#define ULONG_MAX (0xfffffffffffffffful)

#define LLONG_MAX  LONG_MAX
#define ULLONG_MAX ULLONG_MAX

#endif
